/******************************************************************************
 * Copyright (c) 1998 by Mindbright Technology AB, Stockholm, Sweden.
 *
 *                        All Rights Reserved
 *
 *               www.mindbright.se, info@mindbright.se
 *
 * This program and its source-code may be used FREE of charge for strictly
 * NON-COMMERCIAL purposes. Companies or persons wishing to make non-commercial
 * software that in whole or in part contains or is derived from this program
 * or any part thereof must inform the original author of such work. The
 * original author must also be given attribution as the author of used parts.
 * For COMMERCIAL use all parts of this program originally developed by
 * Mindbright Technology must be LICENSED under Mindbright Technology's
 * SHAREWARE license, please contact Mindbright Technology for information on
 * this. This copyright notice must allways be included and unchanged.
 *
 * IF LICENSED FREE OF CHARGE AND/OR IF NOT STATED OTHERWISE THIS PROGRAM COMES
 * WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, IN NO EVENT SHALL
 * MINDBRIGHT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.math.BigInteger;

import mindbright.security.*;
import mindbright.util.*;

public class SSHPduInputStream extends DataInputStream implements SSHPdu {

  static class PduByteArrayInputStream extends ByteArrayInputStream {
    PduByteArrayInputStream(byte[] data) {
      super(data);
    }
    public int getPos() {
      return pos;
    }
    public void setPos(int pos) {
      this.pos = pos;
    }
    public byte[] getBuf() {
      return buf;
    }
    public void setBuf(byte[] buf) {
      this.buf = buf;
    }
  }

  public int type;
  public int length;

  byte[] bytes;
  Cipher cipher;

  SSHPduInputStream(int type, Cipher cipher) {
    super(null);
    this.type   = type; // This is the expected type (checked in readFrom())
    this.cipher = cipher;
  }

  boolean validChecksum() throws IOException {
    int padLen = ((length + 8) & ~7);
    int stored, calculated;

    skip(padLen - 4);
    stored = readInt();
    reset();

    calculated = (int)CRC32.getValue(bytes, 0, padLen - 4);

    if(calculated != stored)
      return false;

    return true;
  }

  public BigInteger readBigInteger() throws IOException {
    short  bits = readShort();
    byte[] raw  = new byte[(bits + 7) / 8 + 1];

    raw[0] = 0;
    read(raw, 1, raw.length - 1);

    return new BigInteger(raw);
  }

  public String readString() throws IOException {
    int    len = readInt();
    byte[] raw = new byte[len];
    read(raw, 0, raw.length);
    return new String(raw);
  }

  public SSHPdu createPdu() {
    return new SSHPduInputStream(this.type, this.cipher);
  }

  public void readFrom(InputStream in) throws IOException {
    DataInputStream dIn    = new DataInputStream(in);
    int             len    = dIn.readInt();
    int             padLen = ((len + 8) & ~7);
    int             type;
    byte[]          data   = new byte[padLen];

    dIn.readFully(data);
    if(cipher != null)
      data = cipher.decrypt(data);

    this.in     = new PduByteArrayInputStream(data);
    this.bytes  = data;
    this.length = len;

    if(!this.validChecksum())
      throw new IOException("Invalid checksum in packet");

    this.skip(8 - (len % 8));
    type = (int)this.readByte();

    if((this.type != SSH.MSG_ANY) && (this.type != type)) {
      this.type = type;
      throw new IOException("Invalid type: " + type + " (expected: " +
			    this.type + ")");
    }
    this.type = type;
  }

  public void writeTo(OutputStream sshOut) throws IOException {
    int off, n;

    if(type != SSH.MSG_CHANNEL_DATA &&
       type != SSH.SMSG_STDOUT_DATA &&
       type != SSH.SMSG_STDERR_DATA)
      throw new IOException("Trying to write raw data from non-data PDU");

    // Here we assume that the content left is readable through readString
    // which is the case if this is SSH-data
    //
    int len = readInt();

    PduByteArrayInputStream is = (PduByteArrayInputStream)in;

    sshOut.write(bytes, is.getPos(), len);

    sshOut.flush();
  }

  public byte[] rawData() {
    PduByteArrayInputStream is = (PduByteArrayInputStream)in;
    return bytes;
  }
  public void rawSetData(byte[] raw) {
    PduByteArrayInputStream is = (PduByteArrayInputStream)in;
    bytes = new byte[raw.length + 4];
    is.setPos(0);
    int len = raw.length;
    int off = 0;
    bytes[off++] = (byte)((len >> 24) & 0xff);
    bytes[off++] = (byte)((len >> 16) & 0xff);
    bytes[off++] = (byte)((len >> 8)  & 0xff);
    bytes[off++] = (byte)(len & 0xff);
    System.arraycopy(raw, 0, bytes, off, raw.length);
    is.setBuf(bytes);
  }
  public int rawOffset() {
    PduByteArrayInputStream is = (PduByteArrayInputStream)in;
    return is.getPos() + 4; // The first four bytes is the length of the data
  }
  public int rawSize() {
    PduByteArrayInputStream is = (PduByteArrayInputStream)in;
    int off = is.getPos();
    int ch1 = bytes[off++];
    int ch2 = bytes[off++];
    int ch3 = bytes[off++];
    int ch4 = bytes[off];
    return ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0));
  }
  public void rawAdjustSize(int size) {
    // !!!
  }

}
