/******************************************************************************
 * Copyright (c) 1998 by Mindbright Technology AB, Stockholm, Sweden.
 *
 *                        All Rights Reserved
 *
 *               www.mindbright.se, info@mindbright.se
 *
 * This program and its source-code may be used FREE of charge for strictly
 * NON-COMMERCIAL purposes. Companies or persons wishing to make non-commercial
 * software that in whole or in part contains or is derived from this program
 * or any part thereof must inform the original author of such work. The
 * original author must also be given attribution as the author of used parts.
 * For COMMERCIAL use all parts of this program originally developed by
 * Mindbright Technology must be LICENSED under Mindbright Technology's
 * SHAREWARE license, please contact Mindbright Technology for information on
 * this. This copyright notice must allways be included and unchanged.
 *
 * IF LICENSED FREE OF CHARGE AND/OR IF NOT STATED OTHERWISE THIS PROGRAM COMES
 * WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, IN NO EVENT SHALL
 * MINDBRIGHT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.awt.*;

import mindbright.terminal.*;
import mindbright.security.*;

public class SSHStdIO implements TerminalListener {

  static public class CtrlDPressedException extends IOException {
  }

  static public class DumbConsoleThread implements Runnable {
    SSHChannelController controller;
    SSHClient            client;
    SSHStdIO             console;

    public DumbConsoleThread(SSHChannelController controller, SSHStdIO console) {
      this.controller = controller;
      this.console    = console;
    }

    public void run() {
      SSHPduOutputStream stdinPdu;
      String line;
      try {
	while(true) {
	  line = console.promptLine("", "", false);
	  stdinPdu = new SSHPduOutputStream(SSH.CMSG_STDIN_DATA, console.sndCipher);
	  stdinPdu.writeString(line + "\n");
	  controller.transmit(stdinPdu);
	  Thread.sleep(400);
	}
      } catch (CtrlDPressedException e) {
	controller.sendDisconnect("exit");
      } catch (Exception e) {
	// !!!
	System.out.println("Error in console-thread: " + e.toString());
      }
    }
  }

  Frame                ownerFrame;
  SSHChannelController controller;
  SSHClient            client;
  Terminal             term;
  Cipher               sndCipher;
  String               ownerName;

  String               scrollbarDef;

  Thread  dumbConsoleThread;

  Boolean readLineLock;
  boolean pressedCtrlD;
  boolean isReadingLine;
  boolean echoStar;
  String  readLineStr;

  protected boolean isConnected;

  public SSHStdIO() {
    this.readLineLock = new Boolean(false);
    this.controller   = null;
    this.sndCipher    = null;
    this.isConnected  = false;

    this.scrollbarDef = "right";
  }

  public boolean isConnected() {
    return isConnected;
  }

  public boolean isDumb() {
    return (term == null);
  }

  public void startDumbConsole() {
    Runnable dumbConsole = new DumbConsoleThread(controller, this);
    dumbConsoleThread = new Thread(dumbConsole);
    dumbConsoleThread.start();
  }
  public void stopDumbConsole() {
    dumbConsoleThread.stop();
  }

  public void serverConnect(SSHChannelController controller, Cipher sndCipher) {
    this.controller  = controller;
    this.sndCipher   = sndCipher;
    isConnected      = true;
  }

  public void serverDisconnect(String reason) {
    this.controller  = null;
    this.sndCipher   = null;
    isConnected      = false;
    println(reason);
  }

  public void setTerminal(Terminal term) {
    this.term = term;
    if(term != null)
      term.addTerminalListener(this);
  }

  public Terminal getTerminal() {
    return term;
  }

  public void setClient(SSHClient client) {
    this.client = client;
  }

  public void setOwnerFrame(Frame ownerFrame) {
    this.ownerFrame = ownerFrame;
  }

  public void setOwnerName(String ownerName) {
    this.ownerName = ownerName;
  }

  public String readLine(String defaultVal) {
    synchronized(readLineLock) {
      if(defaultVal != null) {
	readLineStr   = defaultVal;
	term.write(defaultVal);
      } else {
	readLineStr   = "";
      }
      isReadingLine = true;
      try {
	readLineLock.wait();
      } catch (InterruptedException e) {
	// !!!
      }
      isReadingLine = false;
    }
    return readLineStr;
  }

  public String promptLine(String prompt, String defaultVal, boolean echoStar) throws IOException {
    String line = null;
    pressedCtrlD = false;
    if(term != null) {
      term.setBold(true);
      term.write(prompt);
      term.setBold(false);
      this.echoStar = echoStar;
      line = readLine(defaultVal);
      this.echoStar = false;
    } else {
      BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
      System.out.print(prompt);
      line = br.readLine();

      // !!! kludge, winlost doesn't give empty input with ^D, this is a kludgy thing
      //  anyway since we're not sure that ^D is EOF in the shell...
      //
      if(line == null || line.charAt(0) == (char)4)
	pressedCtrlD = true;
    }
    if(pressedCtrlD) {
      pressedCtrlD = false;
      throw new CtrlDPressedException();
    }
    return line;
  }

  public void print(String str) {
    if(term != null) {
      term.write(str);
    } else {
      System.out.print(str);
    }
  }

  public void println(String str) {
    if(term != null) {
      term.write(str + "\n\r");
    } else {
      System.out.println(str);
    }
  }

  public void updateTitle() {
    int rows = 0;
    int cols = 0;

    if(term == null || ownerFrame == null)
      return;

    rows = term.rows();
    cols = term.cols();

    String title = ownerName + " - ";
    if(client.getUsername() != null)
      title += client.getUsername();
    if(client.getServerName() != null) {
      title += "@" + client.getServerName();
      if(client.getServerVersion() != null)
	title += " <" + client.getServerVersion() + ">";
      else
	title += " <not connected>";
    } else
      title += " <server not known>";
    title += " [" + cols + "x" + rows + "]";
    ownerFrame.setTitle(title);
  }

  // One hell of a kludge all this...
  //
  public String getNextArg(String args) {
    int i = args.indexOf(' ');
    if(i > -1)
      args = args.substring(0, i);
    return args;
  }
  public String[] makeArgv(String cmdLine) {
    String[] argv = new String[32];
    String[] argvRet;
    int n = 0, i;
    while(cmdLine != null) {
      argv[n++] = getNextArg(cmdLine);
      i = cmdLine.indexOf(' ');
      if(i > -1) {
	cmdLine = cmdLine.substring(i);
	cmdLine = cmdLine.trim();
      } else
	cmdLine = null;
    }
    argvRet = new String[n];
    System.arraycopy(argv, 0, argvRet, 0, n);
    return argvRet;
  }
  public void doHelp() {
    println("go                                    Start SSH-session with current settings.");
    println("quit                                  Quit program.");
    println("add <l|r> [/<plug>/]<port>:<host>:<port>  (see below).");
    println("del <l|r> <listen-port>               Remove forwarded local or remote port.");
    println("list                                  Lists current parameters and forwards.");
    println("set <parameter> <value>               Set value of a parameter.");
    println("help                                  Display this list, but you knew that :-).");
    println("");
    println("Examples of adding a remote/local tunnel:");
    println("> add r 4711:www.foo.com:80           Adds a remote tunnel at port 4711 back to");
    println("                                      www.foo.com port 80 without a plugin,");
    println("                                      i.e. default tunneling behaviour.");
    println("> add l /ftp/4711:ftp.foobar.com:21   Adds a local tunnel going to ftp.bar.com");
    println("                                      port 21 using the ftp protocol-plugin to");
    println("                                      handle protocol specific needs.");
    println("");
  }
  public void doAdd(String[] argv) {
    if(argv.length < 3 || (!argv[1].equals("l") && !argv[1].equals("r")))
      doHelp();
    else {
      try {
	if(argv[1].equals("l"))
	  client.addLocalPortForward(argv[2]);
	else
	  client.addRemotePortForward(argv[2]);
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public void doDel(String[] argv) {
    if(argv.length < 3 || (!argv[1].equals("l") && !argv[1].equals("r")))
      doHelp();
    else {
      try {
	if(argv[1].equals("l"))
	  client.delLocalPortForward(Integer.parseInt(argv[2]));
	else
	  client.delRemotePortForward(Integer.parseInt(argv[2]));
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public void doList() {
    int i;
    println("");
    if(term != null)
      term.setBold(true);
    println("SSH settings:");
    if(term != null)
      term.setBold(false);
    println("server   : " + client.getServerName());
    println("port     : " + client.getPort());
    println("username : " + client.getUsername());
    println("local tunnels:");
    for(i = 0; i < client.localForwards.size(); i++) {
      SSHClient.LocalForward fwd = (SSHClient.LocalForward) client.localForwards.elementAt(i);
      println("\tlocal:  " + fwd.localPort + "\tremote: " + fwd.remoteHost + "/" +
	      fwd.remotePort + " (plugin: " + fwd.plugin + ")");
    }
    if(i == 0)
      println("\t<none>");
    println("remote tunnels:");
    for(i = 0; i < client.remoteForwards.size(); i++) {
      SSHClient.RemoteForward fwd = (SSHClient.RemoteForward) client.remoteForwards.elementAt(i);
      println("\tremote: " + fwd.remotePort + "\tlocal:  " + fwd.localHost + "/" +
	      fwd.localPort + " (plugin: " + fwd.plugin + ")");
    }
    if(i == 0)
      println("\t<none>");
    println("");
    if(term != null) {
      TerminalWin termwin = (TerminalWin)term;
      term.setBold(true);
      println("Terminal settings (use 'set' to change):");
      term.setBold(false);
      println("terminal font (fn)           : " + termwin.getFont().getName() + " (" + fontList() + ")");
      println("terminal font size (fs)      : " + termwin.getFont().getSize() + " (system dep.)");
      println("save lines (sl)              : " + termwin.getSaveLines() + " (0 - 32k)");
      println("scroll bar (sb)              : " + scrollbarDef + " (none/left/right)");
      println("reverse video (rv)           : " + termwin.getReverseVideo() + " (true/false)");
      println("auto wrap (aw)               : " + termwin.getAutoWrap() + " (true/false)");
      println("scroll to end on output (si) : " + termwin.getAutoRepositionOnOutput() + " (true/false)");
      println("scroll to end on key (sk)    : " + termwin.getAutoRepositionOnInput() + " (true/false)");
      println("(parameter names and permitted values in parentheses)");
      println("");
    }
  }
  public String fontList() {
    Toolkit tk = ownerFrame.getToolkit();
    String[] fonts = tk.getFontList();
    String list = "";
    for(int i = 0; i < fonts.length; i++) {
      list += fonts[i];
      if(i < fonts.length - 1)
	list += ", ";
    }
    return list;
  }
  public void doSet(String[] argv) {
    if(argv.length < 3)
      doHelp();
    else {
      TerminalWin termwin = (TerminalWin)term;
      try {
	String prm = argv[1];
	String arg = argv[2];
	if(prm.equals("port")) {
	  client.setPort(Integer.parseInt(arg));
	} else if(prm.equals("sl")) {
	  int sl = Integer.parseInt(arg);
	  termwin.setSaveLines(sl);
	} else if(prm.equals("sb")) {
	  if(scrollbarDef.equals(arg))
	    return;
	  if(arg.equals("left") || arg.equals("right")) {
	    scrollbarDef = arg;
	    Scrollbar sb = termwin.getScrollbar();
	    if(sb != null) {
	      ownerFrame.remove(sb);
	    }
	    if(arg.equals("right"))
	      ownerFrame.add(termwin.createScrollbar(), "East");
	    else
	      ownerFrame.add(termwin.createScrollbar(), "West");
	    ownerFrame.pack();
	  } else if(arg.equals("none")) {
	    scrollbarDef = arg;
	    Scrollbar sb = termwin.getScrollbar();
	    if(sb != null) 
	      ownerFrame.remove(sb);
	    ownerFrame.pack();
	  }
	} else if(prm.equals("fn")) {
	  termwin.setFont(arg, termwin.getFont().getSize());
	} else if(prm.equals("fs")) {
	  termwin.setFont(termwin.getFont().getName(), Integer.parseInt(arg));
	} else if(prm.equals("rv")) {
	  termwin.setReverseVideo((new Boolean(arg)).booleanValue());
	} else if(prm.equals("aw")) {
	  termwin.setAutoWrap((new Boolean(arg)).booleanValue());
	} else if(prm.equals("sk")) {
	  termwin.setAutoRepositionOnInput((new Boolean(arg)).booleanValue());
	} else if(prm.equals("si")) {
	  termwin.setAutoRepositionOnOutput((new Boolean(arg)).booleanValue());
	} else if(prm.equals("server")) {
	  client.setServerName(arg);
	} else
	  doHelp();
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public boolean doCommandShell() {
    boolean retVal = true;
    try {
      boolean keepRunning = true;
      String[] argv;
      while(keepRunning) {
	String cmdLine = promptLine("mindterm> ", null, false);
	String cmd;
	cmdLine = cmdLine.toLowerCase();
	cmdLine = cmdLine.trim();
	if(cmdLine.equals(""))
	  continue;
	argv = makeArgv(cmdLine);
	cmd  = argv[0];
	if(cmd.equals("l") || cmd.equals("list")) {
	  doList();
	} else if(cmd.equals("a") || cmd.equals("add")) {
	  doAdd(argv);
	} else if(cmd.equals("d") || cmd.equals("del")) {
	  doDel(argv);
	} else if(cmd.equals("s") || cmd.equals("set")) {
	  doSet(argv);
	} else if(cmd.equals("help") || cmd.equals("?") || cmd.equals("h")) {
	  doHelp();
	} else if(cmd.equals("go") || cmd.equals("g")) {
	  retVal      = true;
	  keepRunning = false;
	} else if(cmd.equals("q") || cmd.equals("quit")) {
	  retVal      = false;
	  keepRunning = false;
	} else {
	  doHelp();
	}
      }
    } catch (Exception e) {
      retVal = false;
    }
    return retVal;
  }

  // TerminalListenerInterface
  //
  public void typedChar(char c) {
    if(isConnected) {
      try {
	SSHPduOutputStream stdinPdu;
	stdinPdu = new SSHPduOutputStream(SSH.CMSG_STDIN_DATA, sndCipher);
	stdinPdu.writeString(String.valueOf(c));
	controller.transmit(stdinPdu);
      } catch (Exception ex) {
	System.out.println("Exception: " + ex.toString());
      }
    } else {
      synchronized(readLineLock) {
	if(isReadingLine) {
	  if(c == (char)4) {
	    // Pressed ctrl-d
	    pressedCtrlD = true;
	    readLineLock.notify();
	  } else if(c == (char)127) {
	    if(readLineStr.length() > 0) {
	      readLineStr = readLineStr.substring(0, readLineStr.length() - 1);
	      term.write((char)8);
	      term.write(' ');
	      term.write((char)8);
	    } else
	      term.doBell();
	  } else if(c == '\r') {
	    //	    readLineStr = readLineStr + "\r";
	    readLineLock.notify();
	    term.write("\n\r");
	  } else {
	    readLineStr = readLineStr + c;
	    if(echoStar)
	      term.write('*');
	    else
	      term.write(c);
	  }
	}
      }
    }
  }

  public void sendBytes(byte[] b) {
    if(isConnected) {
      try {
	SSHPduOutputStream stdinPdu;
	stdinPdu = new SSHPduOutputStream(SSH.CMSG_STDIN_DATA, sndCipher);
	stdinPdu.writeInt(b.length);
	stdinPdu.write(b, 0, b.length);
	controller.transmit(stdinPdu);
      } catch (Exception ex) {
	System.out.println("Exception: " + ex.toString());
      }
    } else {
    }
  }

  public void signalWindowChanged(int rows, int cols, int vpixels, int hpixels) {
    if(isConnected) {
      try {
	SSHPduOutputStream pdu;
	pdu = new SSHPduOutputStream(SSH.CMSG_WINDOW_SIZE, sndCipher);
	pdu.writeInt(rows);
	pdu.writeInt(cols);
	pdu.writeInt(vpixels);
	pdu.writeInt(hpixels);
	controller.transmit(pdu);
      } catch (Exception ex) {
	System.out.println("Exception: " + ex.toString());
      }
    } else {
      // !!!
    }
    updateTitle();
  }

}
